# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import shutil
import subprocess
import sys
import sysconfig
from pathlib import Path
from typing import Optional

from .core import VirtualEnvManager
from .utils import run_subprocess_with_logger

UV_PATH = os.getenv("XOSCAR_UV_PATH")
logger = logging.getLogger(__name__)


def _is_in_pyinstaller():
    return hasattr(sys, "_MEIPASS")


class UVVirtualEnvManager(VirtualEnvManager):
    def __init__(self, env_path: Path):
        super().__init__(env_path)
        self._install_process: Optional[subprocess.Popen] = None

    @classmethod
    def is_available(cls):
        if UV_PATH is not None:
            # user specified uv, just treat it as existed
            return True
        return shutil.which("uv") is not None

    def create_env(self, python_path: Path | None = None) -> None:
        if (uv_path := UV_PATH) is None:
            try:
                from uv import find_uv_bin

                uv_path = find_uv_bin()
            except (ImportError, FileNotFoundError):
                logger.warning("Fail to find uv bin, use system one")
                uv_path = "uv"
        cmd = [uv_path, "venv", str(self.env_path), "--system-site-packages"]
        if python_path:
            cmd += ["--python", str(python_path)]
        elif _is_in_pyinstaller():
            # in pyinstaller, uv would find the system python
            # in this case we'd better specify the same python version
            python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
            cmd += ["--python", python_version]

        logger.info("Creating virtualenv via command: %s", cmd)
        subprocess.run(cmd, check=True)

    def install_packages(self, packages: list[str], **kwargs):
        """
        Install packages into the virtual environment using uv.
        Supports pip-compatible kwargs: index_url, extra_index_url, find_links.
        """
        if not packages:
            return

        # extend the ability of pip
        # maybe replace #system_torch# to the real version
        packages = self.process_packages(packages)
        log = kwargs.pop("log", False)

        uv_path = UV_PATH or "uv"
        cmd = [
            uv_path,
            "pip",
            "install",
            "-p",
            str(self.env_path),
            "--color=always",
        ] + packages

        # Handle known pip-related kwargs
        if "index_url" in kwargs and kwargs["index_url"]:
            cmd += ["-i", kwargs["index_url"]]
        param_and_option = [
            ("extra_index_url", "--extra-index-url"),
            ("find_links", "-f"),
            ("trusted_host", "--trusted-host"),
        ]
        for param, option in param_and_option:
            if param in kwargs and kwargs[param]:
                param_value = kwargs[param]
                if isinstance(param_value, list):
                    for it in param_value:
                        cmd += [option, it]
                else:
                    cmd += [option, param_value]

        logger.info("Installing packages via command: %s", cmd)
        if not log:
            self._install_process = process = subprocess.Popen(cmd)
            returncode = process.wait()
        else:
            with run_subprocess_with_logger(cmd) as process:
                self._install_process = process
            returncode = process.returncode

        self._install_process = None  # install finished, clear reference

        if returncode != 0:
            raise subprocess.CalledProcessError(returncode, cmd)

    def cancel_install(self):
        if self._install_process and self._install_process.poll() is None:
            self._install_process.terminate()
            self._install_process.wait()

    def get_python_path(self) -> str | None:
        if self.env_path.exists():
            return str(self.env_path.joinpath("bin/python"))
        return None

    def get_lib_path(self) -> str:
        return sysconfig.get_path("purelib", vars={"base": str(self.env_path)})

    def remove_env(self):
        if self.env_path.exists():
            shutil.rmtree(self.env_path, ignore_errors=True)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import functools
import inspect
import logging
import threading
import uuid
from collections import defaultdict
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

from .aio import AioFileObject
from .backend import get_backend
from .context import get_context
from .core import ActorRef, BufferRef, FileObjectRef, _Actor, _StatelessActor

if TYPE_CHECKING:
    from .backends.config import ActorPoolConfig
    from .backends.pool import MainActorPoolType

logger = logging.getLogger(__name__)


async def create_actor(
    actor_cls: Type, *args, uid=None, address=None, **kwargs
) -> ActorRef:
    # TODO: explain default values.
    """
    Create an actor.

    Parameters
    ----------
    actor_cls : Actor
        Actor class.
    args : tuple
        Positional arguments for ``actor_cls.__init__``.
    uid : identifier, default=None
        Actor identifier.
    address : str, default=None
        Address to locate the actor.
    kwargs : dict
        Keyword arguments for ``actor_cls.__init__``.

    Returns
    -------
    ActorRef
    """

    ctx = get_context()
    return await ctx.create_actor(actor_cls, *args, uid=uid, address=address, **kwargs)


async def has_actor(actor_ref: ActorRef) -> bool:
    """
    Check if the given actor exists.

    Parameters
    ----------
    actor_ref : ActorRef
        Reference to an actor.

    Returns
    -------
    bool
    """
    ctx = get_context()
    return await ctx.has_actor(actor_ref)


async def destroy_actor(actor_ref: ActorRef):
    """
    Destroy an actor by its reference.

    Parameters
    ----------
    actor_ref : ActorRef
        Reference to an actor.

    Returns
    -------
    bool
    """
    ctx = get_context()
    return await ctx.destroy_actor(actor_ref)


async def actor_ref(*args, **kwargs) -> ActorRef:
    """
    Create a reference to an actor.

    Returns
    -------
    ActorRef
    """
    # TODO: refine the argument list for better user experience.
    ctx = get_context()
    return await ctx.actor_ref(*args, **kwargs)


async def kill_actor(actor_ref: ActorRef):
    # TODO: explain the meaning of 'kill'
    """
    Forcefully kill an actor.

    It's important to note that this operation is potentially
    dangerous as it may result in the termination of other
    associated actors. Only proceed if you understand the
    potential impact on associated actors and can handle any
    resulting consequences.

    Parameters
    ----------
    actor_ref : ActorRef
        Reference to an actor.

    Returns
    -------
    bool
    """
    ctx = get_context()
    return await ctx.kill_actor(actor_ref)


async def create_actor_pool(
    address: str, n_process: int | None = None, **kwargs
) -> "MainActorPoolType":
    # TODO: explain default values.
    """
    Create an actor pool.

    Parameters
    ----------
    address: str
        Address of the actor pool.
    n_process: Optional[int], default=None
        Number of processes.
    kwargs : dict
        Other keyword arguments for the actor pool.

    Returns
    -------
    MainActorPoolType
    """
    if address is None:
        raise ValueError("address has to be provided")
    if "://" not in address:
        scheme = None
    else:
        scheme = urlparse(address).scheme or None

    return await get_backend(scheme).create_actor_pool(
        address, n_process=n_process, **kwargs
    )


async def wait_for(fut: Awaitable[Any], timeout: int | float | None = None) -> Any:
    # asyncio.wait_for() on Xoscar actor call cannot work as expected,
    # because when time out, the future will be cancelled, but an actor call will catch this error,
    # and send a CancelMessage to the dest pool, if the CancelMessage cannot be processed correctly(e.g. the dest pool hangs),
    # the time out will never happen. Thus this PR added a new API so that no matter the CancelMessage delivered or not,
    # the timeout will happen as expected.
    loop = asyncio.get_running_loop()
    new_fut = loop.create_future()
    task = asyncio.ensure_future(fut)

    def on_done(f: asyncio.Future):
        if new_fut.done():
            return
        if f.cancelled():
            new_fut.cancel()
        elif f.exception():
            new_fut.set_exception(f.exception())  # type: ignore
        else:
            new_fut.set_result(f.result())

    task.add_done_callback(on_done)

    try:
        return await asyncio.wait_for(new_fut, timeout)
    except asyncio.TimeoutError:
        if not task.done():
            try:
                task.cancel()  # Try to cancel without waiting
            except Exception:
                logger.warning("Failed to cancel task", exc_info=True)
        raise


def buffer_ref(address: str, buffer: Any) -> BufferRef:
    """
    Init buffer ref according address and buffer.

    Parameters
    ----------
    address
        The address of the buffer.
    buffer
        CPU / GPU buffer. Need to support for slicing and retrieving the length.

    Returns
    ----------
    BufferRef obj.
    """
    ctx = get_context()
    return ctx.buffer_ref(address, buffer)


def file_object_ref(address: str, fileobj: AioFileObject) -> FileObjectRef:
    """
    Init file object ref according to address and aio file obj.

    Parameters
    ----------
    address
        The address of the file obj.
    fileobj
        Aio file object.

    Returns
    ----------
    FileObjectRef obj.
    """
    ctx = get_context()
    return ctx.file_object_ref(address, fileobj)


async def copy_to(
    local_buffers_or_fileobjs: list,
    remote_refs: List[Union[BufferRef, FileObjectRef]],
    block_size: Optional[int] = None,
):
    """
    Copy data from local buffers to remote buffers or copy local file objects to remote file objects.

    Parameters
    ----------
    local_buffers_or_fileobjs
        Local buffers or file objects.
    remote_refs
        Remote buffer refs or file object refs.
    block_size
        Transfer block size when non-ucx
    """
    ctx = get_context()
    return await ctx.copy_to(local_buffers_or_fileobjs, remote_refs, block_size)


async def wait_actor_pool_recovered(address: str, main_pool_address: str | None = None):
    """
    Wait until the specified actor pool has recovered from failure.

    Parameters
    ----------
    address: str
        Address of the actor pool.
    main_pool_address: Optional[str], default=None
        Address of corresponding main actor pool.

    Returns
    -------
    """
    ctx = get_context()
    return await ctx.wait_actor_pool_recovered(address, main_pool_address)


async def get_pool_config(address: str) -> "ActorPoolConfig":
    """
    Get the configuration of specified actor pool.

    Parameters
    ----------
    address: str
        Address of the actor pool.

    Returns
    -------
    ActorPoolConfig
    """
    ctx = get_context()
    return await ctx.get_pool_config(address)


def setup_cluster(address_to_resources: Dict[str, Dict[str, Number]]):
    scheme_to_address_resources: defaultdict[str | None, dict] = defaultdict(dict)
    for address, resources in address_to_resources.items():
        if address is None:
            raise ValueError("address has to be provided")
        if "://" not in address:
            scheme = None
        else:
            scheme = urlparse(address).scheme or None

        scheme_to_address_resources[scheme][address] = resources
    for scheme, address_resources in scheme_to_address_resources.items():
        get_backend(scheme).get_driver_cls().setup_cluster(address_resources)


T = TypeVar("T")


class IteratorWrapper(Generic[T]):
    def __init__(self, uid: str, actor_addr: str, actor_uid: str):
        self._uid = uid
        self._actor_addr = actor_addr
        self._actor_uid = actor_uid
        self._actor_ref = None
        self._gc_destroy = True

    async def destroy(self):
        if self._actor_ref is None:
            self._actor_ref = await actor_ref(
                address=self._actor_addr, uid=self._actor_uid
            )
        assert self._actor_ref is not None
        return await self._actor_ref.__xoscar_destroy_generator__(self._uid)

    def __del__(self):
        # It's not a good idea to spawn a new thread and join in __del__,
        # but currently it's the only way to GC the generator.
        # TODO(codingl2k1): This __del__ may hangs if the program is exiting.
        if self._gc_destroy:
            thread = threading.Thread(
                target=asyncio.run, args=(self.destroy(),), daemon=True
            )
            thread.start()
            thread.join()

    def __aiter__(self):
        return self

    def __getstate__(self):
        # Transfer gc destroy during serialization.
        state = self.__dict__.copy()
        state["_gc_destroy"] = True
        self._gc_destroy = False
        return state

    async def __anext__(self) -> T:
        if self._actor_ref is None:
            self._actor_ref = await actor_ref(
                address=self._actor_addr, uid=self._actor_uid
            )
        try:
            assert self._actor_ref is not None
            return await self._actor_ref.__xoscar_next__(self._uid)
        except Exception as e:
            if "StopIteration" in str(e):
                raise StopAsyncIteration
            else:
                raise


class AsyncActorMixin:
    @classmethod
    def default_uid(cls):
        return cls.__name__

    def __new__(cls, *args, **kwargs):
        try:
            return _actor_implementation[cls](*args, **kwargs)
        except KeyError:
            return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._generators: Dict[str, IteratorWrapper] = {}

    async def __post_create__(self):
        """
        Method called after actor creation
        """
        return await super().__post_create__()

    async def __pre_destroy__(self):
        """
        Method called before actor destroy
        """
        return await super().__pre_destroy__()

    async def __on_receive__(self, message: Tuple[Any]):
        """
        Handle message from other actors and dispatch them to user methods

        Parameters
        ----------
        message : tuple
            Message shall be (method_name,) + args + (kwargs,)
        """
        return await super().__on_receive__(message)  # type: ignore

    async def __xoscar_next__(self, generator_uid: str) -> Any:
        """
        Iter the next of generator.

        Parameters
        ----------
        generator_uid: str
            The uid of generator

        Returns
        -------
            The next value of generator
        """

        def _wrapper(_gen):
            try:
                return next(_gen)
            except StopIteration:
                return stop

        async def _async_wrapper(_gen):
            try:
                # anext is only available for Python >= 3.10
                return await _gen.__anext__()  # noqa: F821
            except StopAsyncIteration:
                return stop

        if gen := self._generators.get(generator_uid):
            stop = object()
            try:
                if inspect.isgenerator(gen):
                    r = await asyncio.to_thread(_wrapper, gen)
                elif inspect.isasyncgen(gen):
                    r = await asyncio.create_task(_async_wrapper(gen))
                else:
                    raise Exception(
                        f"The generator {generator_uid} should be a generator or an async generator, "
                        f"but a {type(gen)} is got."
                    )
            except Exception as e:
                logger.exception(
                    f"Destroy generator {generator_uid} due to an error encountered."
                )
                await self.__xoscar_destroy_generator__(generator_uid)
                del gen  # Avoid exception hold generator reference.
                raise e
            if r is stop:
                await self.__xoscar_destroy_generator__(generator_uid)
                del gen  # Avoid exception hold generator reference.
                raise Exception("StopIteration")
            else:
                return r
        else:
            raise RuntimeError(f"No iterator with id: {generator_uid}")

    async def __xoscar_destroy_generator__(self, generator_uid: str):
        """
        Destroy the generator.

        Parameters
        ----------
        generator_uid: str
            The uid of generator
        """
        logger.debug("Destroy generator: %s", generator_uid)
        self._generators.pop(generator_uid, None)


def generator(func):
    need_to_thread = not asyncio.iscoroutinefunction(func)

    @functools.wraps(func)
    async def _wrapper(self, *args, **kwargs):
        if need_to_thread:
            r = await asyncio.to_thread(func, self, *args, **kwargs)
        else:
            r = await func(self, *args, **kwargs)
        if inspect.isgenerator(r) or inspect.isasyncgen(r):
            gen_uid = uuid.uuid1().hex
            logger.debug("Create generator: %s", gen_uid)
            self._generators[gen_uid] = r
            return IteratorWrapper(gen_uid, self.address, self.uid)
        else:
            return r

    return _wrapper


class Actor(AsyncActorMixin, _Actor):
    pass


class StatelessActor(AsyncActorMixin, _StatelessActor):
    pass


_actor_implementation: Dict[Type[Actor], Type[Actor]] = dict()


def register_actor_implementation(actor_cls: Type[Actor], impl_cls: Type[Actor]):
    _actor_implementation[actor_cls] = impl_cls


def unregister_actor_implementation(actor_cls: Type[Actor]):
    try:
        del _actor_implementation[actor_cls]
    except KeyError:
        pass



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import inspect
import textwrap
from collections import namedtuple
from dataclasses import dataclass
from typing import Any, Callable

from .core import NO_LOCK_ATTRIBUTE_HINT


def build_args_binder(func, remove_self: bool = True) -> Callable | None:
    try:
        spec = inspect.getfullargspec(func)
    except TypeError:  # pragma: no cover
        return None

    sig_list = list(spec.args)
    args_list = list(spec.args)
    if remove_self:
        args_list = args_list[1:]

    if spec.varargs:
        sig_list.append(f"*{spec.varargs}")
        args_list.append(spec.varargs)
    elif spec.kwonlyargs:
        sig_list.append("*")

    sig_list.extend(spec.kwonlyargs)
    args_list.extend(spec.kwonlyargs)

    if spec.varkw:
        sig_list.append(f"**{spec.varkw}")
        args_list.append(spec.varkw)

    if getattr(func, "__name__").isidentifier():
        ret_func_name = f"{func.__name__}_binder"
        ret_type_name = f"_Args_{func.__name__}"
    else:
        ret_func_name = f"anon_{id(func)}_binder"
        ret_type_name = f"_ArgsAnon_{id(func)}"

    func_str = textwrap.dedent(
        f"""
    def {ret_func_name}({', '.join(sig_list)}):
        return {ret_type_name}({', '.join(args_list)})
    """
    )

    glob_vars = globals().copy()
    glob_vars[ret_type_name] = namedtuple(ret_type_name, args_list)
    loc_vars: dict[str, Any] = dict()
    exec(func_str, glob_vars, loc_vars)
    ext_func = loc_vars[ret_func_name]
    ext_func.__defaults__ = spec.defaults
    ext_func.__kwdefaults__ = spec.kwonlydefaults

    return ext_func


@dataclass
class _DelayedArgument:
    args: tuple
    kwargs: dict


class _ExtensibleCallable:
    func: Callable
    batch_func: Callable | None
    is_async: bool
    has_single_func: bool

    def __call__(self, *args, **kwargs):
        if self.is_async:
            return self._async_call(*args, **kwargs)
        else:
            return self._sync_call(*args, **kwargs)

    async def _async_call(self, *args, **kwargs):
        try:
            if self.has_single_func:
                return await self.func(*args, **kwargs)
        except NotImplementedError:
            self.has_single_func = False

        if self.batch_func is not None:
            ret = await self.batch_func([args], [kwargs])
            return None if ret is None else ret[0]
        raise NotImplementedError

    def _sync_call(self, *args, **kwargs):
        try:
            if self.has_single_func:
                return self.func(*args, **kwargs)
        except NotImplementedError:
            self.has_single_func = False

        if self.batch_func is not None:
            return self.batch_func([args], [kwargs])[0]
        raise NotImplementedError


class _ExtensibleWrapper(_ExtensibleCallable):
    def __init__(
        self,
        func: Callable,
        batch_func: Callable | None = None,
        bind_func: Callable | None = None,
        is_async: bool = False,
    ):
        self.func = func
        self.batch_func = batch_func
        self.bind_func = bind_func
        self.is_async = is_async
        self.has_single_func = True

    @staticmethod
    def delay(*args, **kwargs):
        return _DelayedArgument(args=args, kwargs=kwargs)

    @staticmethod
    def _gen_args_kwargs_list(delays):
        args_list = [delay.args for delay in delays]
        kwargs_list = [delay.kwargs for delay in delays]
        return args_list, kwargs_list

    async def _async_batch(self, args_list, kwargs_list):
        # when there is only one call in batch, calling one-pass method
        # will be more efficient
        if len(args_list) == 0:
            return []
        elif len(args_list) == 1:
            return [await self._async_call(*args_list[0], **kwargs_list[0])]
        elif self.batch_func:
            return await self.batch_func(args_list, kwargs_list)
        else:
            # this function has no batch implementation
            # call it separately
            tasks = [
                asyncio.create_task(self.func(*args, **kwargs))
                for args, kwargs in zip(args_list, kwargs_list)
            ]
            try:
                return await asyncio.gather(*tasks)
            except asyncio.CancelledError:
                _ = [task.cancel() for task in tasks]
                return await asyncio.gather(*tasks)

    def _sync_batch(self, args_list, kwargs_list):
        if len(args_list) == 0:
            return []
        elif self.batch_func:
            return self.batch_func(args_list, kwargs_list)
        else:
            # this function has no batch implementation
            # call it separately
            return [
                self.func(*args, **kwargs)
                for args, kwargs in zip(args_list, kwargs_list)
            ]

    def batch(self, *delays):
        args_list, kwargs_list = self._gen_args_kwargs_list(delays)
        return self.call_with_lists(args_list, kwargs_list)

    def call_with_lists(self, args_list, kwargs_list):
        if self.is_async:
            return self._async_batch(args_list, kwargs_list)
        else:
            return self._sync_batch(args_list, kwargs_list)

    def bind(self, *args, **kwargs):
        if self.bind_func is None:
            raise TypeError(f"bind function not exist for method {self.func.__name__}")
        return self.bind_func(*args, **kwargs)


class _ExtensibleAccessor(_ExtensibleCallable):
    func: Callable
    batch_func: Callable | None

    def __init__(self, func: Callable):
        self.func = func
        self.batch_func = None
        self.bind_func = build_args_binder(func, remove_self=True)
        self.is_async = asyncio.iscoroutinefunction(self.func)
        self.has_single_func = True

    def batch(self, func: Callable):
        self.batch_func = func
        return self

    def __get__(self, instance, owner):
        if instance is None:
            # calling from class
            return self.func

        func = self.func.__get__(instance, owner)
        batch_func = (
            self.batch_func.__get__(instance, owner)
            if self.batch_func is not None
            else None
        )
        bind_func = (
            self.bind_func.__get__(instance, owner)
            if self.bind_func is not None
            else None
        )

        wrapper = _ExtensibleWrapper(
            func, batch_func=batch_func, bind_func=bind_func, is_async=self.is_async
        )

        if (
            getattr(self.func, NO_LOCK_ATTRIBUTE_HINT, None) is True
            or getattr(self.batch_func, NO_LOCK_ATTRIBUTE_HINT, None) is True
        ):
            setattr(wrapper, NO_LOCK_ATTRIBUTE_HINT, True)
        return wrapper


def extensible(func: Callable):
    """
    `extensible` means this func could be functionality extended,
    especially for batch operations.

    Consider remote function calls, each function may have operations
    like opening file, closing file, batching them can help to reduce the cost,
    especially for remote function calls.

    Parameters
    ----------
    func : callable
        Function

    Returns
    -------
    func
    """
    return _ExtensibleAccessor(func)



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import subprocess
import sys
import threading
from contextlib import contextmanager
from typing import BinaryIO, Callable, Iterator, List, Optional, TextIO, Union

logger = logging.getLogger(__name__)

ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")


def clean_ansi(text: str) -> str:
    """Remove ANSI escape sequences from text."""
    return ansi_escape.sub("", text)


def stream_reader(
    stream: BinaryIO, log_func: Callable[[str], None], output_stream: TextIO
) -> None:
    """
    Read from the stream, write to logger, and also write to the terminal.
    """
    for line in iter(stream.readline, b""):
        decoded = line.decode(errors="replace")
        output_stream.write(decoded)
        output_stream.flush()
        log_func(clean_ansi(decoded.rstrip("\n")))


@contextmanager
def run_subprocess_with_logger(
    cmd: Union[str, List[str]], cwd: Optional[str] = None, env: Optional[dict] = None
) -> Iterator[subprocess.Popen]:
    """
    Run a subprocess, redirect stdout to logger.info and stderr to logger.error.
    Returns the Popen object as a context manager.

    :param cmd: Command to execute
    :param kwargs: Additional arguments passed to subprocess.Popen
    :yield: The subprocess.Popen object
    """

    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd,
        env=env,
        bufsize=1,
    )

    threads = [
        threading.Thread(
            target=stream_reader, args=(process.stdout, logger.info, sys.stdout)
        ),
        threading.Thread(
            target=stream_reader, args=(process.stderr, logger.error, sys.stderr)
        ),
    ]
    for t in threads:
        t.start()

    try:
        yield process
    finally:
        process.wait()
        for t in threads:
            t.join()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import Enum
from types import TracebackType
from typing import Any, List, Type

from ..core import ActorRef, BufferRef

DEFAULT_PROTOCOL: int = 0

class MessageType(Enum):
    control = 0
    result = 1
    error = 2
    create_actor = 3
    destroy_actor = 4
    has_actor = 5
    actor_ref = 6
    send = 7
    tell = 8
    cancel = 9
    copy_to_buffers = 10
    copy_to_fileobjs = 11
    forward = 12  # message forwarded to other pool

class ControlMessageType(Enum):
    stop = 0
    restart = 1
    sync_config = 2
    get_config = 3
    wait_pool_recovered = 4
    add_sub_pool_actor = 5
    # indicate that the following data will be used for copy_to
    switch_to_copy_to = 6

class _MessageBase:
    message_type: MessageType
    protocol: int
    message_id: bytes
    message_trace: list
    profiling_context: Any

    def __init__(
        self,
        message_id: bytes | None = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
        profiling_context: Any = None,
    ): ...
    def __repr__(self): ...

class CopyToBuffersMessage(_MessageBase):
    message_type = MessageType.copy_to_buffers

    content: object

    def __int__(
        self,
        message_id: bytes | None = None,
        content: object = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class CopyToFileObjectsMessage(CopyToBuffersMessage):
    message_type = MessageType.copy_to_fileobjs

class ControlMessage(_MessageBase):
    message_type = MessageType.control

    address: str
    control_message_type: ControlMessageType
    content: Any

    def __init__(
        self,
        message_id: bytes | None = None,
        address: str | None = None,
        control_message_type: ControlMessageType | None = None,
        content: Any = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class ResultMessage(_MessageBase):
    message_type = MessageType.result

    result: Any

    def __init__(
        self,
        message_id: bytes | None = None,
        result: Any = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
        profiling_context: Any = None,
    ): ...

class ErrorMessage(_MessageBase):
    message_type = MessageType.error

    address: str
    pid: int
    error_type: Type
    error: BaseException
    traceback: TracebackType

    def __init__(
        self,
        message_id: bytes | None = None,
        address: str | None = None,
        pid: int = -1,
        error_type: Type[BaseException] | None = None,
        error: BaseException | None = None,
        traceback: TracebackType | None = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...
    def as_instanceof_cause(self) -> BaseException: ...

class CreateActorMessage(_MessageBase):
    message_type = MessageType.create_actor

    actor_cls: Type
    actor_id: bytes
    args: tuple
    kwargs: dict
    allocate_strategy: Any
    from_main: bool

    def __init__(
        self,
        message_id: bytes | None = None,
        actor_cls: Type | None = None,
        actor_id: bytes | None = None,
        args: tuple | None = None,
        kwargs: dict | None = None,
        allocate_strategy: Any = None,
        from_main: bool = False,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class DestroyActorMessage(_MessageBase):
    message_type = MessageType.destroy_actor

    actor_ref: ActorRef
    from_main: bool

    def __init__(
        self,
        message_id: bytes | None = None,
        actor_ref: ActorRef = None,
        from_main: bool = False,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class HasActorMessage(_MessageBase):
    message_type = MessageType.has_actor

    actor_ref: ActorRef

    def __init__(
        self,
        message_id: bytes | None = None,
        actor_ref: ActorRef = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class ActorRefMessage(_MessageBase):
    message_type = MessageType.actor_ref

    actor_ref: ActorRef

    def __init__(
        self,
        message_id: bytes | None = None,
        actor_ref: ActorRef = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class SendMessage(_MessageBase):
    message_type = MessageType.send

    actor_ref: ActorRef
    content: Any

    def __init__(
        self,
        message_id: bytes | None = None,
        actor_ref: ActorRef = None,
        content: object = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
        profiling_context: Any = None,
    ): ...

class TellMessage(SendMessage):
    message_type = MessageType.tell

class CancelMessage(_MessageBase):
    message_type = MessageType.cancel

    address: str
    cancel_message_id: bytes

    def __init__(
        self,
        message_id: bytes | None = None,
        address: str | None = None,
        cancel_message_id: bytes | None = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class ForwardMessage(_MessageBase):
    message_type = MessageType.forward

    address: str
    raw_message: _MessageBase

    def __init__(
        self,
        message_id: bytes | None = None,
        address: str | None = None,
        raw_message: _MessageBase | None = None,
        protocol: int = DEFAULT_PROTOCOL,
        message_trace: list | None = None,
    ): ...

class DeserializeMessageFailed(RuntimeError):
    message_id: bytes

    def __init__(self, message_id: bytes): ...
    def __str__(self): ...

def new_message_id() -> bytes: ...



# distutils: language = c++
# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
import enum
import hashlib
import inspect
import sys
from functools import partial, wraps
from typing import Any, Dict, List

if sys.version_info[:2] < (3, 8):  # pragma: no cover
    try:
        import pickle5 as pickle  # nosec  # pylint: disable=import_pickle
    except ImportError:
        import pickle  # nosec  # pylint: disable=import_pickle
else:
    import pickle  # nosec  # pylint: disable=import_pickle

import cloudpickle
import numpy as np
import pandas as pd

from cpython cimport PyObject
from libc.stdint cimport int32_t, int64_t, uint32_t, uint64_t, uintptr_t
from libcpp.unordered_map cimport unordered_map

# resolve pandas pickle compatibility between <1.2 and >=1.3
try:
    from pandas.core.internals import blocks as pd_blocks
    if not hasattr(pd_blocks, "new_block") and hasattr(pd_blocks, "make_block"):
        # register missing func that would cause errors
        pd_blocks.new_block = pd_blocks.make_block
except (ImportError, AttributeError):
    pass

from .._utils import NamedType

from .._utils cimport TypeDispatcher
from .pyfury import get_fury

BUFFER_PICKLE_PROTOCOL = max(pickle.DEFAULT_PROTOCOL, 5)
cdef bint HAS_PICKLE_BUFFER = pickle.HIGHEST_PROTOCOL >= 5

cdef TypeDispatcher _serial_dispatcher = TypeDispatcher()
cdef dict _deserializers = dict()

cdef uint32_t _MAX_STR_PRIMITIVE_LEN = 1024
# prime modulus for serializer ids
# use the largest prime number smaller than 32767
cdef int32_t _SERIALIZER_ID_PRIME = 32749


cdef class Serializer:
    serializer_id = None

    def __cinit__(self):
        # make the value can be referenced with C code
        self._serializer_id = self.serializer_id

    cpdef serial(self, object obj, dict context):
        """
        Returns intermediate serialization result of certain object.
        The returned value can be a Placeholder or a tuple comprising
        of three parts: a header, a group of subcomponents and
        a finalizing flag.

        * Header is a pickle-serializable tuple
        * Subcomponents are parts or buffers for iterative
          serialization.
        * Flag is a boolean value. If true, subcomponents should be
          buffers (for instance, bytes, memory views, GPU buffers,
          etc.) that can be read and written directly. If false,
          subcomponents will be serialized iteratively.

        Parameters
        ----------
        obj: Any
            Object to serialize
        context: Dict
            Serialization context to help creating Placeholder objects
            for reducing duplicated serialization

        Returns
        -------
        result: Placeholder | Tuple[Tuple, List, bool]
            Intermediate result of serialization
        """
        raise NotImplementedError

    cpdef deserial(self, tuple serialized, dict context, list subs):
        """
        Returns deserialized object given serialized headers and
        deserialized subcomponents.

        Parameters
        ----------
        serialized: Tuple
            Serialized object header as a tuple
        context
            Serialization context for instantiation of Placeholder
            objects
        subs: List
            Deserialized subcomponents

        Returns
        -------
        result: Any
            Deserialized objects
        """
        raise NotImplementedError

    cpdef on_deserial_error(
        self,
        tuple serialized,
        dict context,
        list subs_serialized,
        int error_index,
        object exc,
    ):
        """
        Returns rewritten exception when subcomponent deserialization fails

        Parameters
        ----------
        serialized: Tuple
            Serialized object header as a tuple
        context
            Serialization context for instantiation of Placeholder
            objects
        subs_serialized: List
            Serialized subcomponents
        error_index: int
            Index of subcomponent causing error
        exc: BaseException
            Exception raised

        Returns
        -------
        exc: BaseException | None
            Rewritten exception. If None, original exception is kept.
        """
        return None

    @classmethod
    def calc_default_serializer_id(cls):
        s = f"{cls.__module__}.{cls.__qualname__}"
        h = hashlib.md5(s.encode())
        return int(h.hexdigest(), 16) % _SERIALIZER_ID_PRIME

    @classmethod
    def register(cls, obj_type, name=None):
        if (
            cls.serializer_id is None
            or cls.serializer_id == getattr(super(cls, cls), "serializer_id", None)
        ):
            # a class should have its own serializer_id
            # inherited serializer_id not acceptable
            cls.serializer_id = cls.calc_default_serializer_id()

        inst = cls()
        if name is not None:
            obj_type = NamedType(name, obj_type)
        _serial_dispatcher.register(obj_type, inst)
        if _deserializers.get(cls.serializer_id) is not None:
            assert type(_deserializers[cls.serializer_id]) is cls
        else:
            _deserializers[cls.serializer_id] = inst

    @classmethod
    def unregister(cls, obj_type, name=None):
        if name is not None:
            obj_type = NamedType(name, obj_type)
        _serial_dispatcher.unregister(obj_type)
        _deserializers.pop(cls.serializer_id, None)


cdef inline uint64_t _fast_id(object obj) nogil:
    return <uintptr_t><PyObject*>obj


def fast_id(obj):
    """C version of id() used for serialization"""
    return _fast_id(obj)


def buffered(func):
    """
    Wrapper for serial() method to reduce duplicated serialization
    """
    @wraps(func)
    def wrapped(self, obj: Any, dict context):
        cdef uint64_t obj_id = _fast_id(obj)
        if obj_id in context:
            return Placeholder(_fast_id(obj))
        else:
            context[obj_id] = obj
            return func(self, obj, context)

    return wrapped


def pickle_buffers(obj):
    cdef list buffers = [None]

    fury = get_fury()
    if fury is not None:
        def buffer_cb(x):
            try:
                buffers.append(memoryview(x))
            except TypeError:
                buffers.append(x.to_buffer())

        buffers[0] = b"__fury__"
        buffers.append(None)
        buffers[1] = fury.serialize(
            obj,
            buffer_callback=buffer_cb,
        )
    else:
        if HAS_PICKLE_BUFFER:
            def buffer_cb(x):
                x = x.raw()
                if x.ndim > 1:
                    # ravel n-d memoryview
                    x = x.cast(x.format)
                buffers.append(memoryview(x))

            buffers[0] = cloudpickle.dumps(
                obj,
                buffer_callback=buffer_cb,
                protocol=BUFFER_PICKLE_PROTOCOL,
            )
        else:
            buffers[0] = cloudpickle.dumps(obj)
    return buffers


def unpickle_buffers(list buffers):
    if buffers[0] == b"__fury__":
        fury = get_fury()
        if fury is None:
            raise Exception("fury is not installed.")
        result = fury.deserialize(buffers[1], buffers[2:])
    else:
        result = cloudpickle.loads(buffers[0], buffers=buffers[1:])

    return result


cdef class PickleSerializer(Serializer):
    serializer_id = 0

    cpdef serial(self, obj: Any, dict context):
        cdef uint64_t obj_id
        obj_id = _fast_id(obj)
        if obj_id in context:
            return Placeholder(obj_id)
        context[obj_id] = obj

        return (), pickle_buffers(obj), True

    cpdef deserial(self, tuple serialized, dict context, list subs):
        return unpickle_buffers(subs)


cdef set _primitive_types = {
    type(None),
    bool,
    int,
    float,
    complex,
    datetime.datetime,
    datetime.date,
    datetime.timedelta,
    enum.Enum,
    type(max),  # builtin functions
    np.dtype,
    np.number,
}


class PrimitiveSerializer(Serializer):
    serializer_id = 1

    @buffered
    def serial(self, obj: Any, context: Dict):
        return (obj,), [], True

    def deserial(self, tuple obj, context: Dict, subs: List[Any]):
        return obj[0]


cdef class BytesSerializer(Serializer):
    serializer_id = 2

    cpdef serial(self, obj: Any, dict context):
        cdef uint64_t obj_id
        obj_id = _fast_id(obj)
        if obj_id in context:
            return Placeholder(obj_id)
        context[obj_id] = obj

        return (), [obj], True

    cpdef deserial(self, tuple serialized, dict context, list subs):
        return subs[0]


cdef class StrSerializer(Serializer):
    serializer_id = 3

    cpdef serial(self, obj: Any, dict context):
        cdef uint64_t obj_id
        obj_id = _fast_id(obj)
        if obj_id in context:
            return Placeholder(obj_id)
        context[obj_id] = obj

        return (), [(<str>obj).encode()], True

    cpdef deserial(self, tuple serialized, dict context, list subs):
        buffer = subs[0]
        if type(buffer) is memoryview:
            buffer = buffer.tobytes()
        return buffer.decode()


cdef class CollectionSerializer(Serializer):
    obj_type = None

    cdef object _obj_type

    def __cinit__(self):
        # make the value can be referenced with C code
        self._obj_type = self.obj_type

    cdef tuple _serial_iterable(self, obj: Any):
        cdef list idx_to_propagate = []
        cdef list obj_to_propagate = []
        cdef list obj_list = <list>obj if type(obj) is list else list(obj)
        cdef int64_t idx
        cdef object item

        for idx in range(len(obj_list)):
            item = obj_list[idx]

            if type(item) is bytes and len(<bytes>item) < _MAX_STR_PRIMITIVE_LEN:
                # treat short strings as primitives
                continue
            elif type(item) is str and len(<str>item) < _MAX_STR_PRIMITIVE_LEN:
                # treat short strings as primitives
                continue
            elif type(item) in _primitive_types:
                continue

            if obj is obj_list:
                obj_list = list(obj)

            obj_list[idx] = None
            idx_to_propagate.append(idx)
            obj_to_propagate.append(item)

        if self._obj_type is not None and type(obj) is not self._obj_type:
            obj_type = type(obj)
        else:
            obj_type = None
        return (obj_list, idx_to_propagate, obj_type), obj_to_propagate, False

    cpdef serial(self, obj: Any, dict context):
        cdef uint64_t obj_id
        obj_id = _fast_id(obj)
        if obj_id in context:
            return Placeholder(obj_id)
        context[obj_id] = obj

        return self._serial_iterable(obj)

    cdef list _deserial_iterable(self, tuple serialized, list subs):
        cdef list res_list, idx_to_propagate
        cdef int64_t i

        res_list, idx_to_propagate, _ = serialized

        for i in range(len(idx_to_propagate)):
            res_list[idx_to_propagate[i]] = subs[i]
        return res_list


cdef class TupleSerializer(CollectionSerializer):
    serializer_id = 4
    obj_type = tuple

    cpdef deserial(self, tuple serialized, dict context, list subs):
        cdef list res = self._deserial_iterable(serialized, subs)
        for v in res:
            assert type(v) is not Placeholder

        obj_type = serialized[-1] or tuple
        if hasattr(obj_type, "_fields"):
            # namedtuple
            return obj_type(*res)
        else:
            return obj_type(res)


cdef class ListSerializer(CollectionSerializer):
    serializer_id = 5
    obj_type = list

    cpdef deserial(self, tuple serialized, dict context, list subs):
        cdef int64_t idx
        cdef list res = self._deserial_iterable(serialized, subs)

        obj_type = serialized[-1]
        if obj_type is None:
            result = res
        else:
            result = obj_type(res)

        for idx, v in enumerate(res):
            if type(v) is Placeholder:
                cb = partial(result.__setitem__, idx)
                (<Placeholder>v).callbacks.append(cb)
        return result


def _dict_key_replacer(ret, key, real_key):
    ret[real_key] = ret.pop(key)


def _dict_value_replacer(context, ret, key, real_value):
    if type(key) is Placeholder:
        key = context[(<Placeholder>key).id]
    ret[key] = real_value


cdef class DictSerializer(CollectionSerializer):
    serializer_id = 6
    cdef set _inspected_inherits

    def __cinit__(self):
        self._inspected_inherits = set()

    cpdef serial(self, obj: Any, dict context):
        cdef uint64_t obj_id
        cdef tuple key_obj, value_obj
        cdef list key_bufs, value_bufs

        if type(obj) is dict and len(<dict>obj) == 0:
            return (), [], True

        obj_id = _fast_id(obj)
        if obj_id in context:
            return Placeholder(obj_id)
        context[obj_id] = obj

        obj_type = type(obj)

        if obj_type is not dict and obj_type not in self._inspected_inherits:
            inspect_init = inspect.getfullargspec(obj_type.__init__)
            if (
                inspect_init.args == ["self"]
                and not inspect_init.varargs
                and not inspect_init.varkw
            ):
                # inherited dicts may not have proper initializers
                # for deserialization
                # remove context to generate real serialized result
                context.pop(obj_id)
                return (obj,), [], True
            else:
                self._inspected_inherits.add(obj_type)

        key_obj, key_bufs, _ = self._serial_iterable(obj.keys())
        value_obj, value_bufs, _ = self._serial_iterable(obj.values())
        if obj_type is dict:
            obj_type = None
        ser_obj = (key_obj[:-1], value_obj[:-1], len(key_bufs), obj_type)
        return ser_obj, key_bufs + value_bufs, False

    cpdef deserial(self, tuple serialized, dict context, list subs):
        cdef int64_t i, num_key_bufs
        cdef list key_subs, value_subs, keys, values

        if not serialized:
            return {}
        if len(serialized) == 1:
            # serialized directly
            return serialized[0]

        key_serialized, value_serialized, num_key_bufs, obj_type = serialized
        key_subs = subs[:num_key_bufs]
        value_subs = subs[num_key_bufs:]

        keys = self._deserial_iterable(<tuple>key_serialized + (None,), key_subs)
        values = self._deserial_iterable(<tuple>value_serialized + (None,), value_subs)

        if obj_type is None:
            ret = dict(zip(keys, values))
        else:
            try:
                ret = obj_type(zip(keys, values))
            except TypeError:
                # first arg of defaultdict is a callable
                ret = obj_type()
                ret.update(zip(keys, values))

        for i in range(len(keys)):
            k, v = keys[i], values[i]
            if type(k) is Placeholder:
                (<Placeholder>k).callbacks.append(
                    partial(_dict_key_replacer, ret, k)
                )
            if type(v) is Placeholder:
                (<Placeholder>v).callbacks.append(
                    partial(_dict_value_replacer, context, ret, k)
                )
        return ret


cdef class Placeholder:
    """
    Placeholder object to reduce duplicated serialization

    The object records object identifier and keeps callbacks
    to replace itself in parent objects.
    """
    cdef public uint64_t id
    cdef public list callbacks

    def __init__(self, uint64_t id_):
        self.id = id_
        self.callbacks = []

    def __hash__(self):
        return self.id

    def __eq__(self, other):  # pragma: no cover
        if type(other) is not Placeholder:
            return False
        return self.id == other.id

    def __repr__(self):
        return (
            f"Placeholder(id={self.id}, "
            f"callbacks=[list of {len(self.callbacks)}])"
        )


cdef class PlaceholderSerializer(Serializer):
    serializer_id = 7

    cpdef serial(self, obj: Any, dict context):
        return (), [], True

    cpdef deserial(self, tuple serialized, dict context, list subs):
        return Placeholder(0)


PickleSerializer.register(object)
for _primitive in _primitive_types:
    PrimitiveSerializer.register(_primitive)
BytesSerializer.register(bytes)
BytesSerializer.register(memoryview)
StrSerializer.register(str)
ListSerializer.register(list)
TupleSerializer.register(tuple)
DictSerializer.register(dict)
PlaceholderSerializer.register(Placeholder)


cdef class _SerialStackItem:
    cdef public tuple serialized
    cdef public list subs
    cdef public list subs_serialized

    def __cinit__(self, tuple serialized, list subs):
        self.serialized = serialized
        self.subs = subs
        self.subs_serialized = []


cdef class _IdContextHolder:
    cdef unordered_map[uint64_t, uint64_t] d


cdef int _COMMON_HEADER_LEN = 4


cdef tuple _serial_single(
    obj, dict context, _IdContextHolder id_context_holder
):
    """Serialize single object and return serialized tuples"""
    cdef uint64_t obj_id, ordered_id
    cdef Serializer serializer
    cdef tuple common_header, serialized

    while True:
        name = context.get("serializer")
        obj_type = type(obj) if name is None else NamedType(name, type(obj))
        serializer = _serial_dispatcher.get_handler(obj_type)
        ret_serial = serializer.serial(obj, context)
        if type(ret_serial) is tuple:
            # object is serialized, form a common header and return
            serialized, subs, final = <tuple>ret_serial

            if type(obj) is Placeholder:
                obj_id = (<Placeholder>obj).id
                ordered_id = id_context_holder.d[obj_id]
            else:
                obj_id = _fast_id(obj)
                ordered_id = id_context_holder.d.size()
                id_context_holder.d[obj_id] = ordered_id

            # REMEMBER to change _COMMON_HEADER_LEN when content of
            # this header changed
            common_header = (
                serializer._serializer_id, ordered_id, len(subs), final
            )
            break
        else:
            # object is converted into another (usually a Placeholder)
            obj = ret_serial
    return common_header + serialized, subs, final


class _SerializeObjectOverflow(Exception):
    def __init__(self, tuple cur_serialized, int num_total_serialized):
        super(_SerializeObjectOverflow, self).__init__(cur_serialized)
        self.cur_serialized = cur_serialized
        self.num_total_serialized = num_total_serialized


cpdef object _serialize_with_stack(
    list serial_stack,
    tuple serialized,
    dict context,
    _IdContextHolder id_context_holder,
    list result_bufs_list,
    int64_t num_overflow = 0,
    int64_t num_total_serialized = 0,
):
    cdef _SerialStackItem stack_item
    cdef list subs
    cdef bint final
    cdef int64_t num_sub_serialized
    cdef bint is_resume = num_total_serialized > 0

    while serial_stack:
        stack_item = serial_stack[-1]
        if serialized is not None:
            # have previously-serialized results, record first
            stack_item.subs_serialized.append(serialized)

        num_sub_serialized = len(stack_item.subs_serialized)
        if len(stack_item.subs) == num_sub_serialized:
            # all subcomponents serialized, serialization of current is done
            # and we can move to the parent object
            serialized = stack_item.serialized + tuple(stack_item.subs_serialized)
            num_total_serialized += 1
            serial_stack.pop()
        else:
            # serialize next subcomponent at stack top
            serialized, subs, final = _serial_single(
                stack_item.subs[num_sub_serialized], context, id_context_holder
            )
            num_total_serialized += 1
            if final or not subs:
                # the subcomponent is a leaf
                if subs:
                    result_bufs_list.extend(subs)
            else:
                # the subcomponent has its own subcomponents, we push itself
                # into stack and process its children
                stack_item = _SerialStackItem(serialized, subs)
                serial_stack.append(stack_item)
                # note that the serialized header should not be recorded
                # as we are now processing the subcomponent itself
                serialized = None
        if 0 < num_overflow < num_total_serialized:
            raise _SerializeObjectOverflow(serialized, num_total_serialized)

    # we keep an empty dict for extra metas required for other modules
    if is_resume:
        # returns num of deserialized objects when resumed
        extra_meta = {"_N": num_total_serialized}
    else:
        # otherwise does not record the number to reduce result size
        extra_meta = {}
    return (extra_meta, serialized), result_bufs_list


def serialize(obj, dict context = None):
    """
    Serialize an object and return a header and buffers.
    Buffers are intended for zero-copy data manipulation.

    Parameters
    ----------
    obj: Any
        Object to serialize
    context:
        Serialization context for instantiation of Placeholder
        objects

    Returns
    -------
    result: Tuple[Tuple, List]
        Picklable header and buffers
    """
    cdef list serial_stack = []
    cdef list result_bufs_list = []
    cdef tuple serialized
    cdef list subs
    cdef bint final
    cdef _IdContextHolder id_context_holder = _IdContextHolder()

    context = context if context is not None else dict()
    serialized, subs, final = _serial_single(obj, context, id_context_holder)
    if final or not subs:
        # marked as a leaf node, return directly
        return ({}, serialized), subs

    serial_stack.append(_SerialStackItem(serialized, subs))
    return _serialize_with_stack(
        serial_stack, None, context, id_context_holder, result_bufs_list
    )


async def serialize_with_spawn(
    obj, dict context = None, int spawn_threshold = 100, object executor = None
):
    """
    Serialize an object and return a header and buffers.
    Buffers are intended for zero-copy data manipulation.

    Parameters
    ----------
    obj: Any
        Object to serialize
    context: Dict
        Serialization context for instantiation of Placeholder
        objects
    spawn_threshold: int
        Threshold to spawn into a ThreadPoolExecutor
    executor: ThreadPoolExecutor
        ThreadPoolExecutor to spawn rest serialization into

    Returns
    -------
    result: Tuple[Tuple, List]
        Picklable header and buffers
    """
    cdef list serial_stack = []
    cdef list result_bufs_list = []
    cdef tuple serialized
    cdef list subs
    cdef bint final
    cdef _IdContextHolder id_context_holder = _IdContextHolder()

    context = context if context is not None else dict()
    serialized, subs, final = _serial_single(obj, context, id_context_holder)
    if final or not subs:
        # marked as a leaf node, return directly
        return ({}, serialized), subs

    serial_stack.append(_SerialStackItem(serialized, subs))

    try:
        result = _serialize_with_stack(
            serial_stack, None, context, id_context_holder, result_bufs_list, spawn_threshold
        )
    except _SerializeObjectOverflow as ex:
        result = await asyncio.get_running_loop().run_in_executor(
            executor,
            _serialize_with_stack,
            serial_stack,
            ex.cur_serialized,
            context,
            id_context_holder,
            result_bufs_list,
            0,
            ex.num_total_serialized,
        )
    return result


cdef class _DeserialStackItem:
    cdef public tuple serialized
    cdef public tuple subs
    cdef public list subs_deserialized

    def __cinit__(self, tuple serialized, tuple subs):
        self.serialized = serialized
        self.subs = subs
        self.subs_deserialized = []


cdef _deserial_single(tuple serialized, dict context, list subs):
    """Deserialize a single object"""
    cdef Serializer serializer
    cdef int64_t num_subs

    serializer_id, obj_id, num_subs, final = serialized[:_COMMON_HEADER_LEN]
    serializer = _deserializers[serializer_id]
    res = serializer.deserial(serialized[_COMMON_HEADER_LEN:], context, subs)

    if type(res) is Placeholder:
        try:
            res = context[obj_id]
        except KeyError:
            (<Placeholder>res).id = obj_id

    # get previously-recorded context values
    context_val, context[obj_id] = context.get(obj_id), res
    # if previously recorded object is a Placeholder,
    # replace it with callbacks
    if type(context_val) is Placeholder:
        for cb in (<Placeholder>context_val).callbacks:
            cb(res)
    return res


def deserialize(tuple serialized, list buffers, dict context = None):
    """
    Deserialize an object with serialized headers and buffers

    Parameters
    ----------
    serialized: Tuple
        Serialized object header
    buffers: List
        List of buffers extracted from serialize() calls
    context: Dict
        Serialization context for replacing Placeholder
        objects

    Returns
    -------
    result: Any
        Deserialized object
    """
    cdef list deserial_stack = []
    cdef _DeserialStackItem stack_item
    cdef int64_t num_subs, num_deserialized, buf_pos = 0
    cdef bint final
    cdef Serializer serializer
    cdef object deserialized = None, exc_value = None
    cdef bint has_deserialized = False

    context = context if context is not None else dict()
    # drop extra meta field
    serialized = serialized[-1]
    serializer_id, obj_id, num_subs, final = serialized[:4]
    if final or num_subs == 0:
        # marked as a leaf node, return directly
        return _deserial_single(serialized, context, buffers)

    deserial_stack.append(
        _DeserialStackItem(
            serialized[:-num_subs], serialized[-num_subs:]
        )
    )

    while deserial_stack:
        stack_item = deserial_stack[-1]
        # the deserialized result can be None, hence we cannot
        # simply judge from the value deserialized
        if has_deserialized:
            # have previously-deserialized results, record first
            stack_item.subs_deserialized.append(deserialized)
        elif exc_value is not None:
            # have exception in successor components, try rewrite
            # and pass to predecessors
            serializer_id = stack_item.serialized[0]
            serializer = _deserializers[serializer_id]
            new_exc_value = serializer.on_deserial_error(
                stack_item.serialized[_COMMON_HEADER_LEN:],
                context,
                list(stack_item.subs),
                len(stack_item.subs_deserialized),
                exc_value,
            )
            exc_value = new_exc_value if new_exc_value is not None else exc_value
            deserial_stack.pop()
            continue

        num_deserialized = len(stack_item.subs_deserialized)
        if len(stack_item.subs) == num_deserialized:
            try:
                # all subcomponents deserialized, we can deserialize the object itself
                deserialized = _deserial_single(
                    stack_item.serialized, context, stack_item.subs_deserialized
                )
                has_deserialized = True
                deserial_stack.pop()
            except BaseException as ex:
                has_deserialized = False
                exc_value = ex
                deserial_stack.pop()
        else:
            # select next subcomponent to process
            serialized = stack_item.subs[num_deserialized]
            serializer_id, obj_id, num_subs, final = serialized[:4]
            if final or num_subs == 0:
                try:
                    # next subcomponent is a leaf, just deserialize
                    deserialized = _deserial_single(
                        serialized, context, buffers[buf_pos : buf_pos + num_subs]
                    )
                    has_deserialized = True
                    buf_pos += num_subs
                except BaseException as ex:
                    has_deserialized = False
                    exc_value = ex
            else:
                # next subcomponent has its own subcomponents, we push it
                # into stack and start handling its children
                stack_item = _DeserialStackItem(
                    serialized[:-num_subs], serialized[-num_subs:]
                )
                deserial_stack.append(stack_item)
                # note that the deserialized object should be cleaned
                # as we are just starting to handle the subcomponent itself
                has_deserialized = False

    if exc_value is not None:
        raise exc_value
    return deserialized



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from pathlib import Path

from .core import VirtualEnvManager
from .uv import UVVirtualEnvManager

_name_to_managers = {"uv": UVVirtualEnvManager}


def get_virtual_env_manager(env_name: str, env_path: str | Path) -> VirtualEnvManager:
    try:
        manager_cls = _name_to_managers[env_name]
    except KeyError:
        raise ValueError(
            f"Unknown virtualenv manager {env_name}, available: {list(_name_to_managers)}"
        )

    path = Path(env_path)
    return manager_cls(path)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Type
from urllib.parse import urlparse

from .base import Client, Server

_scheme_to_client_types: dict[str, Type[Client]] = dict()
_scheme_to_server_types: dict[str, Type[Server]] = dict()


def register_client(client_type: Type[Client]):
    _scheme_to_client_types[client_type.scheme] = client_type  # type: ignore
    return client_type


def register_server(server_type: Type[Server]):
    _scheme_to_server_types[server_type.scheme] = server_type  # type: ignore
    return server_type


def _check_scheme(scheme: str | None, types: dict):
    if scheme == "":
        scheme = None
    if scheme not in types:  # pragma: no cover
        raise ValueError(
            f"address illegal, address scheme "
            f"should be one of "
            f'{", ".join(types)}, '
            f"got {scheme}"
        )
    return scheme


def get_scheme(address: str) -> str | None:
    if "://" not in address:
        scheme = None
    else:
        scheme = urlparse(address).scheme
    return scheme


def get_client_type(address: str) -> Type[Client]:
    scheme = _check_scheme(get_scheme(address), _scheme_to_client_types)
    return _scheme_to_client_types[scheme]


def get_server_type(address: str) -> Type[Server]:
    scheme = _check_scheme(get_scheme(address), _scheme_to_server_types)
    return _scheme_to_server_types[scheme]


def gen_local_address(process_index: int) -> str:
    return f"dummy://{process_index}"



if __name__ == "__main__":
    import click

    @click.group(
        invoke_without_command=True,
        name="xoscar",
        help="Xoscar command-line interface.",
    )
    def main():
        pass

    @main.command("start_sub_pool", help="Start a sub pool.")
    @click.option("shm_name", "-sn", type=str, help="Shared memory name.")
    def start_sub_pool(shm_name):
        from xoscar.backends.indigen.pool import MainActorPool

        MainActorPool._start_sub_pool_in_child(shm_name)

    main()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, Type

from .context import register_backend_context
from .driver import register_backend_driver

__all__ = ["BaseActorBackend", "register_backend", "get_backend"]


class BaseActorBackend(ABC):
    @staticmethod
    @abstractmethod
    def name():
        pass

    @staticmethod
    @abstractmethod
    def get_context_cls():
        pass

    @classmethod
    async def create_actor_pool(
        cls, address: str, n_process: int | None = None, **kwargs
    ):
        pass

    @staticmethod
    @abstractmethod
    def get_driver_cls():
        pass


_scheme_to_backend_cls: Dict[str, Type[BaseActorBackend]] = dict()


def register_backend(backend_cls: Type[BaseActorBackend]):
    name = backend_cls.name()
    if isinstance(name, (list, tuple)):
        names = name
    else:
        names = [name]
    for name in names:
        _scheme_to_backend_cls[name] = backend_cls
        register_backend_context(name, backend_cls.get_context_cls())
        register_backend_driver(name, backend_cls.get_driver_cls())
    return backend_cls


def get_backend(name):
    return _scheme_to_backend_cls[name]



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import time
import weakref
from contextlib import contextmanager
from enum import Enum
from queue import PriorityQueue
from typing import Any, Callable, NamedTuple

from .backends.console import console_metric
from .backends.metric import AbstractMetric
from .backends.prometheus import prometheus_metric

logger = logging.getLogger(__name__)

_init = False
_metric_backend = "console"
_backends_cls = {
    "console": console_metric,
    "prometheus": prometheus_metric,
}


_metrics_to_be_initialized: weakref.WeakSet = weakref.WeakSet()


def init_metrics(backend="console", config: dict[str, Any] | None = None):
    global _init
    if _init is True:
        return

    backend = backend or "console"
    if backend not in _backends_cls:
        raise NotImplementedError(f"Do not support metric backend {backend}")
    global _metric_backend
    _metric_backend = backend
    if _metric_backend == "prometheus":
        try:
            from prometheus_client import start_http_server

            from ..utils import get_next_port

            port = config.get("port", 0) if config else 0
            port = port or get_next_port()
            start_http_server(port)
            logger.warning(
                "Finished startup prometheus http server and port is %d", port
            )
        except ImportError:
            logger.warning(
                "Failed to start prometheus http server because there is no prometheus_client"
            )
    _init = True
    for m in _metrics_to_be_initialized:
        cls = getattr(_backends_cls[_metric_backend], m.type)
        metric = cls(m.name, m.description, m.tag_keys)
        m.set_metric(metric)
    logger.debug("Finished initialize the metrics of backend: %s.", _metric_backend)


def shutdown_metrics():
    global _metric_backend
    _metric_backend = "console"
    global _init
    _init = False
    logger.debug("Shutdown metrics of backend: %s.", _metric_backend)


class _MetricWrapper(AbstractMetric):
    _metric: AbstractMetric | None
    _log_not_init_error: bool

    def __init__(
        self,
        name: str,
        description: str = "",
        tag_keys: tuple[str, ...] | None = None,
        metric_type: str = "Counter",
    ):
        self._name = name
        self._description = description
        self._tag_keys = tag_keys or tuple()
        self._type = metric_type
        self._metric = None
        self._log_not_init_error = False

    @property
    def type(self):
        return self._type

    @property
    def value(self):
        assert (
            self._metric is not None
        ), "Metric is not initialized, please call `init_metrics()` before using metrics."
        return self._metric.value

    def set_metric(self, metric):
        assert metric is not None, "Argument metric is None, please check it."
        self._metric = metric

    def record(self, value=1, tags: dict[str, str] | None = None):
        if self._metric is not None:
            self._metric.record(value, tags)
        elif not self._log_not_init_error:
            self._log_not_init_error = True
            logger.warning(
                "Metric is not initialized, please call `init_metrics()` before using metrics."
            )


def gen_metric(func):
    def wrapper(name, descriptions: str = "", tag_keys: tuple[str, ...] | None = None):
        if _init is True:
            return func(name, descriptions, tag_keys)
        else:
            logger.info(
                "Metric %s will be initialized when invoking `init_metrics()`.", name
            )
            metric = _MetricWrapper(
                name, descriptions, tag_keys, func.__name__.capitalize()
            )
            _metrics_to_be_initialized.add(metric)
            return metric

    return wrapper


class Metrics:
    """
    A factory to generate different types of metrics.

    Note:
        Counter, Meter and Histogram are not thread safe.

    Examples
    --------
    >>> c1 = counter('counter1', 'A counter')
    >>> c1.record(1)

    >>> c2 = counter('counter2', 'A counter', ('service', 'tenant'))
    >>> c2.record(1, {'service': 'indigen', 'tenant': 'test'})

    >>> g1 = gauge('gauge1')
    >>> g1.record(1)

    >>> m1 = meter('meter1')
    >>> m1.record(1)

    >>> h1 = histogram('histogram1')
    >>> h1.record(1)
    """

    @staticmethod
    @gen_metric
    def counter(name, description: str = "", tag_keys: tuple[str, ...] | None = None):
        logger.debug(
            "Initializing a counter with name: %s, tag keys: %s, backend: %s",
            name,
            tag_keys,
            _metric_backend,
        )
        return _backends_cls[_metric_backend].Counter(name, description, tag_keys)

    @staticmethod
    @gen_metric
    def gauge(name, description: str = "", tag_keys: tuple[str, ...] | None = None):
        logger.debug(
            "Initializing a gauge whose name: %s, tag keys: %s, backend: %s",
            name,
            tag_keys,
            _metric_backend,
        )
        return _backends_cls[_metric_backend].Gauge(name, description, tag_keys)

    @staticmethod
    @gen_metric
    def meter(name, description: str = "", tag_keys: tuple[str, ...] | None = None):
        logger.debug(
            "Initializing a meter whose name: %s, tag keys: %s, backend: %s",
            name,
            tag_keys,
            _metric_backend,
        )
        return _backends_cls[_metric_backend].Meter(name, description, tag_keys)

    @staticmethod
    @gen_metric
    def histogram(name, description: str = "", tag_keys: tuple[str, ...] | None = None):
        logger.debug(
            "Initializing a histogram whose name: %s, tag keys: %s, backend: %s",
            name,
            tag_keys,
            _metric_backend,
        )
        return _backends_cls[_metric_backend].Histogram(name, description, tag_keys)


class Percentile:
    class PercentileType(Enum):
        P99 = 1
        P95 = 2
        P90 = 3

    _min_heap: PriorityQueue

    def __init__(self, capacity: int, window: int, callback: Callable[[float], None]):
        self._capacity = capacity
        self._window = window
        self._callback = callback
        self._min_heap = PriorityQueue()
        self._cur_num = 0

        if capacity <= 0 or window <= 0:
            raise ValueError(
                f"capacity or window expect to get a positive integer,"
                f"but capacity got: {capacity} and window got: {window}"
            )

    def record_data(self, value):
        store_value = -1 * value
        if self._min_heap.qsize() < self._capacity:
            self._min_heap.put(store_value)
        else:
            top_value = self._min_heap.get_nowait()
            store_value = store_value if top_value < store_value else top_value
            self._min_heap.put(store_value)

        self._cur_num += 1
        if self._cur_num % self._window == 0:
            self._callback(-1 * self._min_heap.get_nowait())
            self._cur_num = 0
            self._min_heap = PriorityQueue()

    @classmethod
    def build_p99(cls, callback: Callable[[float], None], window: int):
        return cls(int(window * 0.01), window, callback)

    @classmethod
    def build_p95(cls, callback: Callable[[float], None], window: int):
        return cls(int(window * 0.05), window, callback)

    @classmethod
    def build_p90(cls, callback: Callable[[float], None], window: int):
        return cls(int(window * 0.1), window, callback)


_percentile_builder = {
    Percentile.PercentileType.P99: Percentile.build_p99,
    Percentile.PercentileType.P95: Percentile.build_p95,
    Percentile.PercentileType.P90: Percentile.build_p90,
}


class PercentileArg(NamedTuple):
    percentile_type: Percentile.PercentileType
    callback: Callable[[float], None]
    window: int


@contextmanager
def record_time_cost_percentile(percentile_args: list[PercentileArg]):
    percentile_list = [
        _percentile_builder[percentile_type](callback, window)
        for percentile_type, callback, window in percentile_args
    ]
    st_time = time.time()

    yield

    cost_time = time.time() - st_time
    for percentile in percentile_list:
        percentile.record_data(cost_time)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from abc import ABC
from typing import Dict, Optional, Tuple

_THRESHOLD = 2000
_RECORDED_INTERVAL_SECS = 1


class AbstractMetric(ABC):
    """Base class of metrics."""

    _type: str | None = None
    _tag_keys: tuple[str] | tuple | None

    def __init__(
        self, name: str, description: str = "", tag_keys: Optional[Tuple[str]] = None
    ):
        assert isinstance(name, str), "Argument name should be a str"
        assert isinstance(description, str), "Argument description should be a str"
        if tag_keys is not None:
            assert isinstance(tag_keys, tuple) and all(
                isinstance(tag, str) for tag in tag_keys
            ), "Argument tag_keys should be a tuple and its elements should be str"
        self._name = name
        self._description = description
        self._tag_keys = tag_keys or tuple()
        self._init()

    @property
    def type(self):
        return self._type

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def tag_keys(self):
        return self._tag_keys

    def _init(self):
        """Some initialization in subclass."""
        pass

    def record(self, value=1, tags: Optional[Dict[str, str]] = None):
        """A public method called by users."""
        pass

    def _record(self, value: float = 1.0, tags: Optional[Dict[str, str]] = None):
        """An internal method called by record() and should be
        implemented by different metric backends.
        """
        pass


class AbstractCounter(AbstractMetric):
    """A counter records the counts of events."""

    _type = "Counter"

    def __init__(
        self, name: str, description: str = "", tag_keys: Optional[Tuple[str]] = None
    ):
        super().__init__(name, description, tag_keys)
        self._count = 0

    def record(self, value=1, tags: Optional[Dict[str, str]] = None):
        self._count += value
        self._record(self._count, tags)


class AbstractGauge(AbstractMetric):
    """A gauge represents a single numerical value that can be
    arbitrarily set.
    """

    _type = "Gauge"

    def record(self, value=1, tags: Optional[Dict[str, str]] = None):
        self._record(value, tags)


class AbstractMeter(AbstractMetric):
    """A meter measures the rate at which a set of events occur."""

    _type = "Meter"

    def __init__(
        self, name: str, description: str = "", tag_keys: Optional[Tuple[str]] = None
    ):
        super().__init__(name, description, tag_keys)
        self._count = 0
        self._last_time = time.time()

    def record(self, value=1, tags: Optional[Dict[str, str]] = None):
        self._count += value
        now = time.time()
        past = max(now - self._last_time, 1e-9)
        if self._count >= _THRESHOLD or past >= _RECORDED_INTERVAL_SECS:
            qps = self._count / past
            self._record(qps, tags)
            self._last_time = now
            self._count = 0


class AbstractHistogram(AbstractMetric):
    """A histogram measures the distribution of values in a stream of data."""

    _type = "Histogram"
    _data: list

    def __init__(
        self, name: str, description: str = "", tag_keys: Optional[Tuple[str]] = None
    ):
        super().__init__(name, description, tag_keys)
        self._data = list()
        self._last_time = time.time()

    def record(self, value=1, tags: Optional[Dict[str, str]] = None):
        self._data.append(value)
        now = time.time()
        if (
            len(self._data) >= _THRESHOLD
            or now - self._last_time >= _RECORDED_INTERVAL_SECS
        ):
            avg = sum(self._data) / len(self._data)
            self._record(avg, tags)
            self._data.clear()
            self._last_time = now



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from ctypes import c_void_p
from enum import IntEnum
from typing import Callable, List, Optional

import xoscar_pygloo

class ReduceOp(IntEnum):
    SUM = 0
    PRODUCT = 1
    MIN = 2
    MAX = 3
    BAND = 4
    BOR = 5
    BXOR = 6
    UNUSED = 7

class GlooDataType_t(IntEnum):
    glooInt8 = 0
    glooUint8 = 1
    glooInt32 = 2
    glooUint32 = 3
    glooInt64 = 4
    glooUint64 = 5
    glooFloat16 = 6
    glooFloat32 = 7
    glooFloat64 = 8

class AllreduceAlgorithm(IntEnum):
    UNSPECIFIED = 0
    RING = 1
    BCUBE = 2

def transport_tcp_available() -> bool: ...
def transport_uv_available() -> bool: ...

class Context:
    rank: Optional[int] = None
    size: Optional[int] = None
    base: Optional[int] = 2
    def getDevice(self) -> int: ...
    def createUnboundBuffer(self, ptr: c_void_p, size: int): ...
    def nextSlot(self, numToSkip: int) -> int: ...
    def closeConnections(self) -> None: ...
    def setTimeout(self, timeout: datetime.timedelta) -> None: ...
    def getTimeout(self) -> datetime.timedelta: ...

def allreduce(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    reduceop: Optional[ReduceOp] = ReduceOp.SUM,
    algorithm: Optional[AllreduceAlgorithm] = AllreduceAlgorithm.RING,
    tag: int = 0,
) -> None: ...
def allgather(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    tag: Optional[int] = 0,
) -> None: ...
def all_to_all(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    tag: Optional[int] = 0,
) -> None: ...
def allgatherv(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    tag: Optional[int] = 0,
) -> None: ...
def reduce(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    reduceop: Optional[ReduceOp] = ReduceOp.SUM,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
) -> None: ...
def scatter(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
) -> None: ...
def gather(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
) -> None: ...
def send(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    peer: Optional[int] = None,
    tag: Optional[int] = 0,
) -> None: ...
def recv(
    context: Optional[Context] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    peer: Optional[int] = None,
    tag: Optional[int] = 0,
) -> None: ...
def broadcast(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    datatype: Optional[GlooDataType_t] = None,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
) -> None: ...
def reduce_scatter(
    context: Optional[Context] = None,
    sendbuf: Optional[int] = None,
    recvbuf: Optional[int] = None,
    size: Optional[int] = None,
    recvElems: Optional[List[int]] = None,
    datatype: Optional[GlooDataType_t] = None,
    reduceop: Optional[ReduceOp] = ReduceOp.SUM,
) -> None: ...
def barrier(context: Optional[Context] = None, tag: Optional[int] = 0) -> None: ...

class rendezvous:
    class Store:
        def set(self, key: str, data: List[str]) -> None: ...
        def get(self, key: str) -> str: ...

    class TCPStoreOptions:
        port: int
        isServer: bool
        numWorkers: Optional[int]
        waitWorkers: bool
        timeout: datetime.timedelta
        multiTenant: bool

    class TCPStore:
        def __init__(
            self,
            host: str,
            opts: rendezvous.TCPStoreOptions = rendezvous.TCPStoreOptions(),
        ): ...
        def set(self, key: str, value: bytes): ...
        def get(self, key: str) -> bytes: ...
        def wait(self, keys: List[str]): ...

    class FileStore(Store):
        def __init__(self, path: str) -> None: ...
        def set(self, key: str, data: List[str]) -> None: ...
        def get(self, key: str) -> str: ...

    class HashStore(Store):
        def __init__(self) -> None: ...
        def set(self, key: str, data: List[str]) -> None: ...
        def get(self, key: str) -> str: ...

    class PrefixStore(Store):
        def __init__(self, prefix: str, store: rendezvous.Store) -> None: ...
        def set(self, key: str, data: List[str]) -> None: ...
        def get(self, key: str) -> str: ...

    class CustomStore(Store):
        def __init__(self, real_store_py_object: object) -> None: ...
        def delKeys(self, keys: List[str]) -> None: ...
        def set(self, key: str, data: List[str]) -> None: ...
        def get(self, key: str) -> str: ...

    class Context(xoscar_pygloo.Context):
        def connectFullMesh(
            self, store: rendezvous.Store, dev: transport.Device
        ) -> None: ...

class transport:
    class uv:
        pass

    class tcp:
        class Device(transport.Device):
            def __init__(self, attr: transport.tcp.attr) -> None: ...

        class Context(xoscar_pygloo.Context):
            def __init__(
                self, device: transport.tcp.Device, rank: int, size: int
            ) -> None: ...

        class attr:
            hostname: str
            iface: str
            ai_family: int
            ai_socktype: int
            ai_protocol: int
            ai_addr: object
            ai_addrlen: int
            def __init__(self, string: Optional[str] = None) -> None: ...

        def CreateDevice(self, src: transport.tcp.attr) -> transport.Device: ...

    class Device:
        def __str__(self) -> str: ...
        def getPCIBusID(self) -> Callable[[], str]: ...
        def getInterfaceSpeed(self) -> int: ...
        def hasGPUDirect(self) -> bool: ...
        def createContext(self, rank: int, size: int): ...



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import numpy as np

from ..utils import lazy_import

cupy = lazy_import("cupy")


def convert_data_to_np_array(data):
    if isinstance(data, np.ndarray):
        return data
    else:
        return np.frombuffer(data, dtype="u1")


def convert_data_to_cp_array(data):
    if isinstance(data, cupy.ndarray):
        return data
    else:
        return cupy.frombuffer(data, dtype="u1")


def get_rank_address_via_env(env_key: str, err_message: str) -> str:
    address = os.environ.get(env_key, None)
    if address is None:
        raise RuntimeError(err_message)
    return address



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


cdef class TypeDispatcher:
    cdef dict _handlers
    cdef dict _lazy_handlers
    cdef dict _inherit_handlers
    cdef object __weakref__

    cpdef void register(self, object type_, object handler)
    cpdef void unregister(self, object type_)
    cdef _reload_lazy_handlers(self)
    cpdef get_handler(self, object type_)

cpdef str to_str(s, encoding=*)
cpdef bytes to_binary(s, encoding=*)
cpdef bytes new_random_id(int byte_len)
cpdef bytes new_actor_id()
cdef bint is_async_generator(obj)


cdef class Timer:
    cdef object _start
    cdef readonly object duration



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
from typing import Any, Optional

from ..communication import DummyServer, gen_local_address
from ..config import ActorPoolConfig
from ..indigen.pool import MainActorPool, SubActorPool
from ..message import ControlMessage, ControlMessageType, new_message_id
from ..pool import ActorPoolType


class TestMainActorPool(MainActorPool):
    @classmethod
    def get_external_addresses(
        cls,
        address: str,
        n_process: int | None = None,
        ports: list[int] | None = None,
        schemes: list[Optional[str]] | None = None,
    ):
        if "://" in address:
            address = address.split("://", 1)[1]
        return super().get_external_addresses(address, n_process=n_process, ports=ports)

    @classmethod
    def gen_internal_address(
        cls, process_index: int, external_address: str | None = None
    ) -> str:
        return f"dummy://{process_index}"

    @classmethod
    async def start_sub_pool(
        cls,
        actor_pool_config: ActorPoolConfig,
        process_index: int,
        start_python: str | None = None,
    ):
        return await cls._create_sub_pool_test(actor_pool_config, process_index, 0)

    @classmethod
    async def wait_sub_pools_ready(cls, create_pool_tasks: list[asyncio.Task]):
        addresses = []
        tasks = []
        for t in create_pool_tasks:
            pool_task, external_addresses = await t
            tasks.append(pool_task)
            addresses.append(external_addresses)
        return tasks, addresses

    @classmethod
    async def _create_sub_pool_test(
        cls,
        actor_config: ActorPoolConfig,
        process_index: int,
        main_pool_pid: int,
    ):
        pool: TestSubActorPool = await TestSubActorPool.create(
            {
                "actor_pool_config": actor_config,
                "process_index": process_index,
                "main_pool_pid": main_pool_pid,
            }
        )
        await pool.start()
        actor_config.reset_pool_external_address(process_index, [pool.external_address])
        cur_pool_config = actor_config.get_pool_config(process_index)
        return None, cur_pool_config["external_address"]

    def _sync_pool_config(self, actor_pool_config: ActorPoolConfig):
        # test pool does not create routers, thus can skip this step
        pass

    async def append_sub_pool(
        self,
        label: str | None = None,
        internal_address: str | None = None,
        external_address: str | None = None,
        env: dict | None = None,
        modules: list[str] | None = None,
        suspend_sigint: bool | None = None,
        use_uvloop: bool | None = None,
        logging_conf: dict | None = None,
        start_python: str | None = None,
        kwargs: dict | None = None,
    ):
        external_address = (
            external_address
            or TestMainActorPool.get_external_addresses(
                self.external_address, n_process=1
            )[-1]
        )

        # use last process index's logging_conf and use_uv_loop config if not provide
        actor_pool_config = self._config.as_dict()
        last_process_index = self._config.get_process_indexes()[-1]
        last_logging_conf = actor_pool_config["pools"][last_process_index][
            "logging_conf"
        ]
        last_use_uv_loop = actor_pool_config["pools"][last_process_index]["use_uvloop"]
        _logging_conf = logging_conf or last_logging_conf
        _use_uv_loop = use_uvloop if use_uvloop is not None else last_use_uv_loop

        process_index = next(TestMainActorPool.process_index_gen(external_address))
        internal_address = internal_address or TestMainActorPool.gen_internal_address(
            process_index, external_address
        )

        self._config.add_pool_conf(
            process_index,
            label,
            internal_address,
            external_address,
            env,
            modules,
            suspend_sigint,
            _use_uv_loop,
            _logging_conf,
            kwargs,
        )
        pool_task = asyncio.create_task(
            TestMainActorPool.start_sub_pool(self._config, process_index)
        )
        tasks, addresses = await TestMainActorPool.wait_sub_pools_ready([pool_task])

        self.attach_sub_process(addresses[0][0], tasks[0])

        control_message = ControlMessage(
            message_id=new_message_id(),
            address=self.external_address,
            control_message_type=ControlMessageType.sync_config,
            content=self._config,
        )
        await self.handle_control_command(control_message)

        return addresses[0][0]

    async def kill_sub_pool(
        self, process: asyncio.subprocess.Process, force: bool = False
    ):
        pass

    async def is_sub_pool_alive(self, process: asyncio.subprocess.Process):
        return True


class TestSubActorPool(SubActorPool):
    def _sync_pool_config(self, actor_pool_config: ActorPoolConfig):
        # test pool does not create routers, thus can skip this step
        pass

    @classmethod
    async def create(cls, config: dict) -> ActorPoolType:
        kw: dict[str, Any] = dict()
        cls._parse_config(config, kw)
        process_index: int = kw["process_index"]
        actor_pool_config = kw["config"]  # type: ActorPoolConfig
        external_addresses = actor_pool_config.get_pool_config(process_index)[
            "external_address"
        ]

        def handle_channel(channel):
            return pool.on_new_channel(channel)

        # create servers
        server_addresses = external_addresses + [gen_local_address(process_index)]
        server_addresses = sorted(set(server_addresses))
        servers = await cls._create_servers(
            server_addresses, handle_channel, actor_pool_config.get_comm_config()
        )
        cls._update_stored_addresses(servers, server_addresses, actor_pool_config, kw)

        # create pool
        pool = cls(**kw)
        return pool  # type: ignore

    async def stop(self):
        # do not close dummy server
        self._servers = [
            s for s in self._servers[:-1] if not isinstance(s, DummyServer)
        ]
        await super().stop()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple

import pandas as pd

from ..utils import lazy_import
from .core import Serializer, buffered

cupy = lazy_import("cupy")
cudf = lazy_import("cudf")


class CupySerializer(Serializer):
    @buffered
    def serial(self, obj: Any, context: Dict):
        if not (obj.flags["C_CONTIGUOUS"] or obj.flags["F_CONTIGUOUS"]):
            obj = cupy.array(obj, copy=True)

        header = obj.__cuda_array_interface__.copy()
        header["strides"] = tuple(obj.strides)
        header["lengths"] = [obj.nbytes]
        buffer = cupy.ndarray(
            shape=(obj.nbytes,), dtype=cupy.dtype("u1"), memptr=obj.data, strides=(1,)
        )
        return (header,), [buffer], True

    def deserial(self, serialized: Tuple, context: Dict, subs: List):
        (header,) = serialized
        return cupy.ndarray(
            shape=header["shape"],
            dtype=header["typestr"],
            memptr=cupy.asarray(subs[0]).data,
            strides=header["strides"],
        )


class CudfSerializer(Serializer):
    @staticmethod
    def _get_ext_index_type(index_obj):
        import cudf

        multi_index_type = None
        if isinstance(index_obj, pd.MultiIndex):
            multi_index_type = "pandas"
        elif isinstance(index_obj, cudf.MultiIndex):
            multi_index_type = "cudf"

        if multi_index_type is None:
            return None
        return {
            "index_type": multi_index_type,
            "names": list(index_obj.names),
        }

    @staticmethod
    def _apply_index_type(obj, attr, header):
        import cudf

        multi_index_cls = (
            pd.MultiIndex if header["index_type"] == "pandas" else cudf.MultiIndex
        )
        original_index = getattr(obj, attr)
        if isinstance(original_index, (pd.MultiIndex, cudf.MultiIndex)):
            return
        new_index = multi_index_cls.from_tuples(original_index, names=header["names"])
        setattr(obj, attr, new_index)

    def serial(self, obj: Any, context: Dict):
        header, buffers = obj.device_serialize()
        if hasattr(obj, "columns"):
            header["_ext_columns"] = self._get_ext_index_type(obj.columns)
        if hasattr(obj, "index"):
            header["_ext_index"] = self._get_ext_index_type(obj.index)
        return (header,), buffers, True

    def deserial(self, serialized: Tuple, context: Dict, buffers: List):
        from cudf.core.abc import Serializable

        (header,) = serialized
        col_header = header.pop("_ext_columns", None)
        index_header = header.pop("_ext_index", None)

        result = Serializable.device_deserialize(header, buffers)

        if col_header is not None:
            self._apply_index_type(result, "columns", col_header)
        if index_header is not None:
            self._apply_index_type(result, "index", index_header)
        return result


if cupy is not None:
    CupySerializer.register("cupy.ndarray")
if cudf is not None:
    CudfSerializer.register("cudf.DataFrame")
    CudfSerializer.register("cudf.Series")
    CudfSerializer.register("cudf.Index")



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import concurrent.futures as futures
import logging
import weakref
from typing import Any, Callable, Coroutine, Dict, Optional, Type
from urllib.parse import urlparse

from ...errors import ServerClosed
from ...utils import abc_type_require_weakref_slot, classproperty, implements
from .base import Channel, ChannelType, Client, Server
from .core import register_client, register_server
from .errors import ChannelClosed

DEFAULT_DUMMY_ADDRESS = "dummy://0"

logger = logging.getLogger(__name__)


class DummyChannel(Channel):
    """
    Channel for communications in same process.
    """

    __slots__ = "__weakref__", "_in_queue", "_out_queue", "_closed"

    name = "dummy"

    def __init__(
        self,
        in_queue: asyncio.Queue,
        out_queue: asyncio.Queue,
        closed: asyncio.Event,
        local_address: str | None = None,
        dest_address: str | None = None,
        compression: str | None = None,
    ):
        super().__init__(
            local_address=local_address,
            dest_address=dest_address,
            compression=compression,
        )
        self._in_queue = in_queue
        self._out_queue = out_queue
        self._closed = closed

    @property
    @implements(Channel.type)
    def type(self) -> int:
        return ChannelType.local

    @implements(Channel.send)
    async def send(self, message: Any):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed, cannot send message")
        # put message directly into queue
        self._out_queue.put_nowait(message)

    @implements(Channel.recv)
    async def recv(self):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed("Channel already closed, cannot write message")
        try:
            return await self._in_queue.get()
        except RuntimeError:
            if self._closed.is_set():
                pass

    @implements(Channel.close)
    async def close(self):
        self._closed.set()

    @property
    @implements(Channel.closed)
    def closed(self) -> bool:
        return self._closed.is_set()


@register_server
class DummyServer(Server):
    __slots__ = (
        ("_closed", "_channels", "_tasks") + ("__weakref__",)
        if abc_type_require_weakref_slot
        else tuple()
    )

    _address_to_instances: weakref.WeakValueDictionary[str, "DummyServer"] = (
        weakref.WeakValueDictionary()
    )
    _channels: weakref.WeakSet[Channel]
    _tasks: set[asyncio.Task]
    scheme: str | None = "dummy"

    def __init__(
        self,
        address: str,
        channel_handler: Callable[[Channel], Coroutine] | None = None,
    ):
        super().__init__(address, channel_handler)
        self._closed = asyncio.Event()
        self._channels = weakref.WeakSet()
        self._tasks = set()

    @classmethod
    def get_instance(cls, address: str):
        return cls._address_to_instances[address]

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return DummyClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> int:
        return ChannelType.local

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "DummyServer":
        config = config.copy()
        address = config.pop("address", DEFAULT_DUMMY_ADDRESS)
        handle_channel = config.pop("handle_channel")
        if urlparse(address).scheme != DummyServer.scheme:  # pragma: no cover
            raise ValueError(
                f"Address for DummyServer "
                f'should be starts with "dummy://", '
                f"got {address}"
            )
        if config:  # pragma: no cover
            raise TypeError(
                f"Creating DummyServer got unexpected " f'arguments: {",".join(config)}'
            )
        try:
            server = DummyServer.get_instance(address)
            if server.stopped:
                raise KeyError("server closed")
        except KeyError:
            server = DummyServer(address, handle_channel)
            DummyServer._address_to_instances[address] = server
        return server

    @implements(Server.start)
    async def start(self):
        # nothing needs to do for dummy server
        pass

    @implements(Server.join)
    async def join(self, timeout=None):
        wait_coro = self._closed.wait()
        try:
            await asyncio.wait_for(wait_coro, timeout=timeout)
        except (futures.TimeoutError, asyncio.TimeoutError):
            pass

    @implements(Server.on_connected)
    async def on_connected(self, *args, **kwargs):
        if self._closed.is_set():  # pragma: no cover
            raise ServerClosed("Dummy server already closed")

        channel = args[0]
        assert isinstance(channel, DummyChannel)
        if kwargs:  # pragma: no cover
            raise TypeError(
                f"{type(self).__name__} got unexpected "
                f'arguments: {",".join(kwargs)}'
            )
        self._channels.add(channel)
        await self.channel_handler(channel)

    @implements(Server.stop)
    async def stop(self):
        self._closed.set()
        _ = [t.cancel() for t in self._tasks]
        await asyncio.gather(*(channel.close() for channel in self._channels))

    @property
    @implements(Server.stopped)
    def stopped(self) -> bool:
        return self._closed.is_set()


@register_client
class DummyClient(Client):
    __slots__ = ("_task",)

    scheme: str | None = DummyServer.scheme

    def __init__(
        self, local_address: str | None, dest_address: str | None, channel: Channel
    ):
        super().__init__(local_address, dest_address, channel)
        self._task: Optional[asyncio.Task] = None

    @staticmethod
    @implements(Client.connect)
    async def connect(
        dest_address: str, local_address: str | None = None, **kwargs
    ) -> "Client":
        if urlparse(dest_address).scheme != DummyServer.scheme:  # pragma: no cover
            raise ValueError(
                f'Destination address should start with "dummy://" '
                f"for DummyClient, got {dest_address}"
            )
        server = DummyServer.get_instance(dest_address)
        if server is None:  # pragma: no cover
            raise RuntimeError(
                f"DummyServer {dest_address} needs to be created first before DummyClient"
            )
        if server.stopped:  # pragma: no cover
            raise ConnectionError(f"Dummy server {dest_address} closed")

        q1: asyncio.Queue = asyncio.Queue()
        q2: asyncio.Queue = asyncio.Queue()
        closed = asyncio.Event()
        client_channel = DummyChannel(q1, q2, closed, local_address=local_address)
        server_channel = DummyChannel(q2, q1, closed, dest_address=local_address)

        conn_coro = server.on_connected(server_channel)
        task = asyncio.create_task(conn_coro)
        client = DummyClient(local_address, dest_address, client_channel)
        client._task = task
        server._tasks.add(task)

        def _discard(t):
            server._tasks.discard(t)
            logger.info("Channel exit: %s", server_channel.info)

        task.add_done_callback(_discard)
        return client

    @implements(Client.close)
    async def close(self):
        await super().close()
        if self._task is not None:
            self._task.cancel()
            self._task = None
